# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.dianomic.com
# FOGLAMP_END


__author__ = "Mohit Rajput, Ashwini Kumar Pandey"
__copyright__ = "Copyright (c) 2025 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


import os
import linecache
import logging
import time
import tracemalloc
from threading import Event, Thread
import psutil

LOG_LEVEL = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARN": logging.WARN,
    "ERROR": logging.ERROR
}

try:
    from foglamp.common import logger
    _LOGGER = logger.setup(__name__, level=logging.INFO)
except:
    class _LOGGER:
        debug = print
        info = print
        warn = print
        error = print

DEFAULT_SCORE = -1
DEFAULT_FORECAST = -1
DEFAULT_FORECAST_STATE_NAME = -1
DEFAULT_FORECAST_STATE_ID = -1


class TrackResourceUsage:
    """Track the CPU and memory usage during code execution.

    This class monitors the following:
        - Memory: Peak and retained memory usage (in MB).
        - CPU: Percentage of CPU usage and CPU time taken (in seconds)

    References:
        https://docs.python.org/3/library/tracemalloc.html#functions
        https://docs.python.org/3/library/tracemalloc.html#pretty-top
        https://stackoverflow.com/questions/552744/how-do-i-profile-memory-usage-in-python
    """

    def __init__(self):
        self.memory_retained_at_end = None
        self.peak_memory_consumption = None

        self.cpu_time_taken = None
        self.start_cpu_time = None
        self.start_wall_time = None
        self.cpu_percent_start = None
        self.cpu_percent_end = None

    def __enter__(self):
        tracemalloc.start()
        tracemalloc.clear_traces()

        self.start_cpu_time = time.process_time()
        self.start_wall_time = time.time()  # Wall-clock time
        self.cpu_percent_start = psutil.cpu_percent(interval=None)

        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.memory_retained_at_end, self.peak_memory_consumption = (
            tracemalloc.get_traced_memory()
        )
        tracemalloc.stop()

        self.cpu_time_taken = time.process_time() - self.start_cpu_time
        self.cpu_percent_end = psutil.cpu_percent(interval=None)

    def stats(self):
        """Return observed peak allocated memory and retained memory by code snippet."""
        if self.memory_retained_at_end is not None:
            self.memory_retained_at_end = self.memory_retained_at_end / 1024 / 1024
            self.peak_memory_consumption = self.peak_memory_consumption / 1024 / 1024

        return {
            "memory_retained_at_end_MB": self.memory_retained_at_end,
            "memory_peak_consumption_MB": self.peak_memory_consumption,
            "cpu_time_taken_sec": self.cpu_time_taken,
            "cpu_percent_start": self.cpu_percent_start / 100,
            "cpu_percent_end": self.cpu_percent_end / 100,
        }


class RepeatedTimer:
    """A robust implementation of a timer that executes a function at fixed intervals."""

    def __init__(self, interval, function, *args, **kwargs):
        """Initialize the RepeatedTimer.

        Args:
            interval (float): Time interval in seconds between function executions.
            function (callable): The function to execute periodically.
            *args: Positional arguments to pass to the function.
            **kwargs: Keyword arguments to pass to the function.
        """
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.stop_event = Event()
        self.thread = Thread(target=self._run, daemon=True)  # Use a daemon thread
        self.is_running = False

    def _run(self):
        """Internal method to repeatedly execute the function at fixed intervals."""
        next_run = time.time()
        while not self.stop_event.is_set():
            next_run += self.interval
            self.function(*self.args, **self.kwargs)
            # Wait for the next interval
            sleep_time = next_run - time.time()
            if sleep_time > 0:
                self.stop_event.wait(sleep_time)

    def start(self):
        """Start the repeated timer in a background thread."""
        if not self.is_running:
            self.is_running = True
            self.stop_event.clear()  # Ensure the stop event is cleared before starting
            self.thread.start()

    def stop(self):
        """Stop the repeated timer and clean up the thread."""
        self.stop_event.set()
        self.is_running = False
        self.thread.join()  # Ensure the thread stops cleanly

    def __del__(self):
        """Ensure the timer is stopped when the object is deleted."""
        self.stop()


class TriggerExecution:
    """A class that help repeatedly indicate if execution is allowed when the configured time has expired."""

    def __init__(self, wait_period=10):
        self.wait_period = wait_period
        self.repeated_timer = RepeatedTimer(wait_period, TriggerExecution._execution_allowed, self)
        self.repeated_timer.start()
        self.execution_allowed = False

    def _execution_allowed(self):
        self.execution_allowed = True

    @property
    def is_execution_allowed(self):
        return self.execution_allowed

    @property
    def reset_timer(self):
        self.execution_allowed = False

    @property
    def shutdown_timer(self):
        self.repeated_timer.stop()
        self.repeated_timer = None
