"""This file contains bucket configuration and code for interacting with Bucket Service."""

__author__ = "Mohit Rajput"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"
__email__ = "info@dianomic.com"


import os
import asyncio
from foglamp.common.bucket_client.model_query_builder import ModelQueryBuilder

bucket_config_properties = {
    "constant": {
        "type": "model"
    },
    "key": {
        "name": {
            "order": "1",
            "displayName": "Model name",
            "description": "Model to be loaded from bucket service.",
            "type": "string",
            "default": "breakover_forecasting_model.tflite"
        },
    },
    "properties": {
        "version": {
            "order": "2",
            "displayName": "Model version",
            "description": "Model version as stored in bucket.",
            "type": "string",
            "default": "1.0"
        }
    }
}

bucket_config_default = {
    "type": bucket_config_properties["constant"]["type"],
    "name": bucket_config_properties["key"]["name"]["default"],
    "version": bucket_config_properties["properties"]["version"]["default"],
}


def get_model_path(
    logger, bucket_client, bucket_type, model_name, model_version
):
    """Get model file from the bucket service.

    FUTURE: load model from buffer without writing to disk.

    Args:
        bucket_client: Bucket client handle to use bucket services
        bucket_type: type defined in the attributes in bucket.
            it should match with
            bucket_config_properties.constant.type --> `model`
        model_name: Model name
            e.g. `breakover_forecasting_model.h5`
        model_version: Model's version
            e.g. `1.0`
    Returns:
        Model path: path where the models has been saved.
    """
    # paths setup
    model_dir = os.path.join("/tmp", "foglamp_breakover_filter")
    model_saving_path = os.path.join(
        model_dir,
        bucket_type,
        model_version.replace(" ", "_").replace(".", "_").lower(),
        model_name  # "breakover_forecasting_model.h5"
    )

    # prepare directory
    os.system(f"rm -rf {model_dir}")
    os.makedirs(os.path.split(model_saving_path)[0])

    # creating query for searching the bucket using the attributes
    model_attrs = {
        "type": bucket_type,
        "name": model_name,
        "version": model_version,
    }
    mqb = ModelQueryBuilder()
    for param, user_defined_value in model_attrs.items():
        mqb = mqb.MATCH([param, "=", user_defined_value])
    query = mqb.payload()
    logger.debug(f"bucket_client.query_models: query payload={query}")

    # create a temporary event loop for async calls in this function since
    # sometimes this function gets called before event loop creation
    # (plugin_init) and sometimes after (plugin_reconfigure)
    loop = asyncio.new_event_loop()

    # locating the matching bucket
    try:
        # getting the matching bucket options based on the query created above
        task = asyncio.ensure_future(
            bucket_client.query_models(query), loop=loop
        )
        loop.run_until_complete(task)
        result = task.result()
        logger.debug(f"bucket_client.query_models returned {result}")

        # checking the matches
        if len(result["matches"]) < 1:
            logger.error("Didn\'t find any matching data within bucket service.")
            return None
        if len(result["matches"]) == 1:
            logger.info("Found the exact match within bucket service. Loading it.")
        if len(result["matches"]) > 1:
            logger.warn("Found more than one match within bucket service. Picking first one.")

        # loading the match
        result = result["matches"][0]
        logger.info(f"Chosen model : {result}")

    except Exception as ex:
        _msg = "Breakover Forecasting model is unable to get a match within bucket service."
        _msg += f"\nDetails: {str(ex)}"
        logger.error(_msg)
        return None

    # get data from the bucket & save it to the mentioned path
    task = asyncio.ensure_future(
        bucket_client.get_bucket_content(result["bucket"], model_saving_path),
        loop=loop
    )
    loop.run_until_complete(task)
    result = task.result()
    loop.stop()
    loop.close()
    logger.debug(f"Model saved from the bucket to `{model_saving_path}`")

    return model_saving_path
